/********************************************************************
* PROJECT: LoonBoard Project
*
* COYRIGHT: Colin O'Flynn of NewAE, (C) 2005 <coflynn@newae.com>
*
* LICENSE: This file is released under a license specified in the LICENSE
*          file in the main project directory that this file was in. If
*          the license file is missing, then this file is NOT licensed
*          and all standard copyright laws apply. 
*
* FILE NAME: lub.c                                                  
* 
* PURPOSE: Provides a number of routines to use the LoonBoard Unified
*          Bootloader (LUB) protocol to transmit and receive packets.
* 
* FILE REFERENCES: 
* 
* Name                I/O   Description 
* ----                ---   ----------- 
* serial_lowlevel.c   IO    lowlevel serial communications for one byte
*
* EXTERNAL VARIABLES: 
* Source: 
* 
* Name Type I/O Description 
* ---- ---- --- ----------- 
* none 
*
* EXTERNAL REFERENCES: 
* 
* Name          Description 
* ----          ----------- 
* serial_getc   Gets one byte from the serial port, returns it
* serial_putc   Puts one byte passed as an argument to the serial port
*
* ABNORMAL TERMINATION CONDITIONS, ERROR AND WARNING MESSAGES: 
* 
* ASSUMPTIONS, CONSTRAINTS, RESTRICTIONS: 
*     Assumes that the serial routines will always work OK, which is
*     normally pretty safe as long as they were opened. The return
*     values are never checked.
* 
* NOTES: 
* 
* REQUIREMENTS/FUNCTIONAL SPECIFICATIONS REFERENCES: 
* 
* DEVELOPMENT HISTORY: 
* 
* Date(DMY)   Author          Release   Description Of Change 
* ----        ------          -------   --------------------- 
* 30-04-2005  Colin O'Flynn   Initial   Initial file written
* 14-05-2005  Colin O'Flynn   0.1       Changes
*
*******************************************************************/

/* Include Files */
#include "serial_lowlevel.h"
#include "crc8.h"
#include "lub.h"

#ifndef NULL
#define NULL 0
#endif


static unsigned char lub_packet_counter = 1;

/* Local function prototypes */
unsigned int lub_NumberBytes        (unsigned char packettype);
int lub_GetPacketStream_lowlevel    (void (*char_process)(unsigned char, unsigned int));
int lub_GetPacket_lowlevel          (char * data, int maxsize) ;
int lub_SendPacket_lowlevel         (unsigned char packettype, unsigned char * data);

/********************************************************
* 
* FUNCTION NAME: lub_GetPacketStream
* 
* ARGUMENTS: 
* 
* ARGUMENT      TYPE         I/O  DESCRIPTION 
* --------      ----         ---  ----------- 
* char_process  ptr to func   I   On every databyte (note:
*                                 data part of packet only)
*                                 received this function will
*                                 be called with two arguments:
*                                 the first is the current
*                                 databyte, and the second
*                                 is the databyte number.
*                                 When the databyte number is
*                                 zero, the data held is the
*                                 packet type
*                                   an example:
*                                  void function_name (
*                                    unsigned char data,
*                                    unsigned int bytenum);
*
* verify_data   ptr to func  I   Once the packet is received
*                                this function will be called
*                                if the packet was OK, and when
*                                this function returns it will
*                                send an ACK
*
* maxtries     unsigned char I   Maximum number of times to try 
*                                before giving up
*
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      All OK, Packet verified OK
*   -1     Ran out of tries
*
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/

int lub_GetPacketStream(void (*char_process)(unsigned char, unsigned int), void (*verify_data)(void), unsigned char maxtries)
    {
            
    lub_packet_counter++;
    
    while (maxtries != 0)
        {
        lub_packet_counter--;
        
        if (lub_GetPacketStream_lowlevel(char_process) == 0)
            {
            //Call user routine
            verify_data();
            
            //Send ACK
            lub_packet_counter--;
            
            lub_SendPacket_lowlevel(LUB_ACK, NULL);
            
            return 0;
            }
        else
            {
            //send NACK
            lub_packet_counter--;
            
            lub_SendPacket_lowlevel(LUB_NACK, NULL);
            }            
        
        maxtries--;
        }
    
    return -1;
    }
    
/********************************************************
* 
* FUNCTION NAME: lub_GetPacket
* 
* ARGUMENTS: 
* 
* ARGUMENT      TYPE         I/O  DESCRIPTION 
* --------      ----         ---  ----------- 
* data          ptr to char   O   Incomming packet stored
*                                 at this location
*
* maxsize       int           I   Max number of bytes to 
*                                 store
*
* maxtries      unsigned char I   Max number of times to try
* 
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      All OK, Packet verified OK
*   -1     Packet ran out of tries
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/

int lub_GetPacket(char * data, int maxsize, unsigned char maxtries) 
    {
 //   fprintf(stderr,"GetPacketStream: Attempting to get packet ");
    
    lub_packet_counter++;
            
    while (maxtries != 0)
        {
        //fprintf(stderr,"%d...", lub_packet_counter);
        lub_packet_counter--;
        
        if (lub_GetPacket_lowlevel(data, maxsize) == 0)
            {
            //Send ACK
            lub_packet_counter--;
            
            lub_SendPacket_lowlevel(LUB_ACK, NULL);
            
            return 0;
            }
        else
            {
            //send NACK
            lub_packet_counter--;
            
            lub_SendPacket_lowlevel(LUB_NACK, NULL);
            }            
        
        maxtries--;
        }
    
    return -1;
    }
    

/********************************************************
* 
* FUNCTION NAME: lub_SendPacket
* 
* ARGUMENTS: 
* 
* ARGUMENT    TYPE            I/O   DESCRIPTION 
* --------    ----            ---   ----------- 
* packettype  unsigned char    I    The LUB packet type
*                                   that this packet is, 
*                                   specified in LUB specs
*
* data        * unsigned char  I    Should point to any
*                                   data that needs to be
*                                   transmitted. The size of
*                                   this buffer will depend
*                                   on the packet type. If
*                                   the sent packet doesn't
*                                   have any data, set to
*                                   NULL
*
* maxtries   unsigned char     I    Maximum number of times to
*                                   try
*
*
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      Packet Sent
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/
int lub_SendPacket(unsigned char packettype, unsigned char * data, unsigned char maxtries)
    {
    unsigned char ack_buffer[10];
    
    lub_packet_counter++;

    
    while (maxtries != 0)
        {
        
        //lub_SendPacket will increment the counter
        lub_packet_counter--;
        lub_SendPacket_lowlevel(packettype, data);
            
        //GetPacket will increment the packet counter
        lub_packet_counter--;
        if (lub_GetPacket_lowlevel(ack_buffer, 10) == 0)
            {
            if (ack_buffer[2] == LUB_ACK)
                {
                return 0;
                }
            }
        maxtries--;
        }
    
    return -1;
    }    

/********************************************************
* 
* FUNCTION NAME: lub_GetPacketStream_lowlevel
* 
* ARGUMENTS: 
* 
* ARGUMENT      TYPE         I/O  DESCRIPTION 
* --------      ----         ---  ----------- 
* char_process  ptr to func   I   On every databyte (note:
*                                 data part of packet only)
*                                 received this function will
*                                 be called with two arguments:
*                                 the first is the current
*                                 databyte, and the second
*                                 is the databyte number.
*                                 When the databyte number is
*                                 zero, the data held is the
*                                 packet type
*                                   an example:
*                                  void function_name (
*                                    unsigned char data,
*                                    unsigned int bytenum);
* 
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      All OK, Packet verified OK
*   -1     Packet NOT OK, CRC8 error or packet number
*          error 
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/

int lub_GetPacketStream_lowlevel(void (*char_process)(unsigned char, unsigned int))
    {    
    unsigned char recv_packet_counter;
    unsigned char recv_packettype;
    
    unsigned int  number_data_bytes;
    unsigned int  bytenumber;
    
    unsigned char errors = 0;
    
    unsigned char c;
    
    //Clear CRC8
    crc8_clear();
    
    //wait for sync
    while (serial_getc() != LUB_SYNC)
        {
        continue;
        }

    //get byte count
    recv_packet_counter = serial_getc();
    crc8_addbyte(recv_packet_counter);
    
    //get packet type
    recv_packettype = serial_getc();
    crc8_addbyte(recv_packettype);
    
    /* If the command is for the start of a packet, this must be packet #1 */
    if (recv_packettype == LUB_SOT)
        {
        lub_packet_counter = 1;
        }
    else
        {
        lub_packet_counter++;
        }
        
    //Check this is the packet we are expecting
    if (lub_packet_counter != recv_packet_counter)
        {
        errors |= 0x80;
        }
  
        
    //Figure out how much data we should be getting
     number_data_bytes = lub_NumberBytes(recv_packettype); 
     
     //Receive as many databytes as we should, processing
     //them as instructed and then CRC8ing them
     bytenumber = 0;
     char_process(recv_packettype, bytenumber);
     
     while (number_data_bytes != 0)
        {
        bytenumber++;
        c = serial_getc();
        crc8_addbyte(c);
        char_process(c, bytenumber);
        
        number_data_bytes--;
        }
        
     //The final step.. CRC8 the CRC8 received, we should get 0
     crc8_addbyte(serial_getc());
     
     if (crc8_getcrc() != 0)
        {
        errors |= 0x01;
        }
        
     if (errors == 0)
        {
        return 0;
        }
     else
        {
        return -1;
        }
     }
     
/********************************************************
* 
* FUNCTION NAME: lub_GetPacket_lowlevel
* 
* ARGUMENTS: 
* 
* ARGUMENT      TYPE         I/O  DESCRIPTION 
* --------      ----         ---  ----------- 
* data          ptr to char   O   Incomming packet stored
*                                 at this location
*
* maxsize       int           I   Max number of bytes to 
*                                 store
* 
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      All OK, Packet verified OK
*   -1     Packet NOT OK, CRC8 error or packet number
*          error 
*   -2     Packet was too big to store
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/

int lub_GetPacket_lowlevel(char * data, int maxsize) 
    {
    
    unsigned char recv_packet_counter;
    unsigned char recv_packettype;
    
    unsigned int  number_data_bytes;
    
    unsigned char errors = 0;
    
    unsigned char c;
    
    //Clear CRC8
    crc8_clear();
    
    //wait for sync
    while (serial_getc() != LUB_SYNC)
        {
        continue;
        }
    *data = LUB_SYNC;
    data++;

    //get byte count
    recv_packet_counter = serial_getc();
    *data = recv_packet_counter;
    data++;
    crc8_addbyte(recv_packet_counter);
    
    //get packet type
    recv_packettype = serial_getc();
    *data = recv_packettype;
    data++;
    crc8_addbyte(recv_packettype);
    
    /* If the command is for the start of a packet, this must be packet #1 */
    if (recv_packettype == LUB_SOT)
        {
        lub_packet_counter = 1;
        }
    else
        {
        lub_packet_counter++;
        }
       
    //Check this is the packet we are expecting
    if (lub_packet_counter != recv_packet_counter)
        {
        errors |= 0x80;
        }
        
    //Figure out how much data we should be getting
     number_data_bytes = lub_NumberBytes(recv_packettype); 
     
     //Check we have enough room
     if ((number_data_bytes + 4) > maxsize)
        {
        errors |= 0x40;
        data = NULL;
        }
     
     //Receive as many databytes as we should, processing
     //them as instructed and then CRC8ing them
     serial_getcs(data, number_data_bytes);
     while (number_data_bytes != 0)
        {
        //c = serial_getc();
        //*data = c;
        crc8_addbyte(*data);
        data++;        
        number_data_bytes--;
        }
        
     //The final step.. CRC8 the CRC8 received, we should get 0
      crc8_addbyte(*data = serial_getc());
     
     if (crc8_getcrc() != 0)
        {
        errors |= 0x01;
        }
        
     if (errors == 0)
        {
        return 0;
        }
     else
        {
        return -1;
        }
     }
        
           
        
        
/********************************************************
* 
* FUNCTION NAME: lub_SendPacket_lowlevel
* 
* ARGUMENTS: 
* 
* ARGUMENT    TYPE            I/O   DESCRIPTION 
* --------    ----            ---   ----------- 
* packettype  unsigned char    I    The LUB packet type
*                                   that this packet is, 
*                                   specified in LUB specs
*
* data        * unsigned char  I    Should point to any
*                                   data that needs to be
*                                   transmitted. The size of
*                                   this buffer will depend
*                                   on the packet type. If
*                                   the sent packet doesn't
*                                   have any data, set to
*                                   NULL
*
*
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      Packet Sent
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/
int lub_SendPacket_lowlevel(unsigned char packettype, unsigned char * data)
    {
    
    unsigned int number_data_bytes;
    
    //Clear CRC8
    crc8_clear();
    
    //Send sync
    serial_putc(LUB_SYNC);
       
    /* If the command is for the start of a packet, this must be packet #1 */
    if (packettype == LUB_SOT)
        {
        lub_packet_counter = 1;
        }
    else
        {
        lub_packet_counter++;
        }
    
    //Send byte counter
    serial_putc(lub_packet_counter);
    crc8_addbyte(lub_packet_counter);
    
    //Send type of packet
    serial_putc(packettype);
    crc8_addbyte(packettype);
    
    //Send data bytes if there should be some
     number_data_bytes = lub_NumberBytes(packettype);

        
    /* For every data byte send it and add to CRC8 */
    serial_putcs(data, number_data_bytes);
    
    while(number_data_bytes != 0)
        {
        //serial_putc(*data);
        crc8_addbyte(*data);
        data++;
        number_data_bytes--;
        }
    
    /* Send CRC8 */
    serial_putc(crc8_getcrc());
         
    return 0;
    }
    
    
/********************************************************
* 
* FUNCTION NAME: lub_EnterLUB
*
* DESCRIPTION: Tries to enter the LUB
* 
* ARGUMENTS: 
* 
* ARGUMENT    TYPE            I/O   DESCRIPTION 
* --------    ----            ---   ----------- 
* which_lub   unsigned char    I    When set to 0 loads the 
*                                   normal LUB, when set to
*                                   1 loads the emergency 
*                                   LUB
*
* RETURNS: int depending on error or success:
* 
* VALUE    MEANING
* -----    --------------- 
*   0      LUB Entered
*   -1     LUB did not enter
*  Others  Oh crap - something VERY wrong
*  
*********************************************************/    
int lub_EnterLUB(unsigned char which_lub)
    {
    unsigned char   lubmode; /* Which lub character to send */
    
    /* Wait for the ? */
    while(serial_getc() != '?')
    {
    continue;  
    }
    
    /* Send a 's' for safe mode bootloader, and a 'b' for the
       normal mode bootloader */
    
    if (which_lub == 0)
        {
        lubmode = 'b'; /* Normal bootloader */
        }
    else if (which_lub == 3)
        {
        lubmode = 'd'; /* Debug */
        }
    else 
        {
        lubmode = 's'; /* Failsafe bootloader */
        }
        
    serial_putc(lubmode);        
    

    /* Sometimes lubloader has to fake it and will send a second ? */    
    /* So we check it twice... */
    if (serial_getc() == 'A')
        {
        return 0;
        }
    if (serial_getc() == 'A')
        {
        return 0;
        }
        
    return -1;
        
    }
    
    
    
/********************************************************
* 
* FUNCTION NAME: lub_NumberBytes
* 
* ARGUMENTS: 
* 
* ARGUMENT    TYPE            I/O   DESCRIPTION 
* --------    ----            ---   ----------- 
* packettype  unsigned char    I    The LUB packet type
*                                   that this packet is, 
*                                   specified in LUB specs
*
* RETURNS: Unsigned int corresponding to the number of data
*          bytes that that type of packet should have attached
* 
*********************************************************/    
unsigned int lub_NumberBytes(unsigned char packettype)
    {
        //Send data bytes if there should be some
    if (packettype == LUB_264_DATA)
        {
        return 264;
        }
    else if (packettype == LUB_SIGN_ON || packettype == LUB_CRC1)
        {
        return 2;
        }
    else
        {
        return 0;
        }
    }


