/********************************************************************
* PROJECT: LoonBoard Project
*
* COYRIGHT: Colin O'Flynn of NewAE, (C) 2005 <coflynn@newae.com>
*
* LICENSE: This file is released under a license specified in the LICENSE
*          file in the main project directory that this file was in. If
*          the license file is missing, then this file is NOT licensed
*          and all standard copyright laws apply. 
*
* FILE NAME: main.c                                        
* 
* PURPOSE: Main file for the Loonboard Unified Bootloader (LUB) that runs
*          on the AVR target
* 
* FILE REFERENCES: 
* 
* Name                I/O   Description 
* ----                ---   ----------- 
* lub.c               IO    LUB Protocol communications routines
* uart.c              IO    Simple UART serial communications
* dataflash.c         IO    Communications for DataFLASH
*
*
* ABNORMAL TERMINATION CONDITIONS, ERROR AND WARNING MESSAGES: 
* 
* ASSUMPTIONS, CONSTRAINTS, RESTRICTIONS: 
* 
* NOTES: 
* 
* REQUIREMENTS/FUNCTIONAL SPECIFICATIONS REFERENCES: 
* 
* DEVELOPMENT HISTORY: 
* 
* Date(DMY)   Author          Release   Description Of Change 
* ----        ------          -------   --------------------- 
* 15-05-2005  Colin O'Flynn    0.1      First working release
*
*******************************************************************/


#include <avr/io.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <stdlib.h>
#include "led_control.h"

#ifndef F_CPU
#error "CPU Frequency Undefined"
#endif //F_CPU

#include "calibrateosc.h"
#include "twi_comms.h"
#include "x1226.h"

//Setup baud rate, BAUDRATE is defined in Makefile
#define ubrr (unsigned int)((F_CPU / (16 * BAUDRATE) ) - 1)

#include "lub.h"
#include "uart.h"
#include "serial_lowlevel.h"
#include "dataflash.h"
#include "crc8.h"

static unsigned int DF_buffernum = 1;
static unsigned int DF_page = 0;
static unsigned char Global_packettype; 
static unsigned char Write_flash = 0;

static unsigned int CRCStartPage = 0;

void ProcessLubData(unsigned char data, unsigned int bytenum);
void FlashData(void);
void DealWithCRC(void);
void CommitSuicide(void);

int main(void)
    {
    unsigned char buffer[10];
  
    /* Enable LEDs */
    LED_INIT();
    
    LED_RED(ON);
    LED_GREEN(ON);
    
     //Enable SPI for read/writing from DataFLASH memory
    DF_SPI_init();
    Read_DF_status();
         
    /* Setup X1226 */
    twi_init();
     
    /* Only do this if X1226 is responding */
    if (x1226_init() == ERROR_NONE)
        {
    
        buffer[0] = 0x02;
        buffer[1] = 0x06;
    
        //Enable writing to X1226 CCR Register
        x1226_random_write(X1226_CCR, 0x3F, &buffer[0], 1);
        x1226_random_write(X1226_CCR, 0x3F, &buffer[1], 1);
    
        //Setup 4096 KHz clock
        DDRD &= ~(1<<3);
        PORTD |= 1<<3;
        buffer[0] = 1<<4;
        x1226_random_write(X1226_CCR, 0x11, &buffer[0], 1);
   
        //Setup clock to 7372800 Hz
        CalOsc();
        }
 
    //Store new OSCCAL value
    Buffer_Write_Byte(1, 0, OSCCAL);
    Buffer_Write_Byte(1, 1, ~OSCCAL); //Stores inverse
    Buffer_To_Page(1, 1);
    while(!(Read_DF_status() & 0x80));

    LED_RED(OFF);
          
    /*Enable receiver and transmitter */ 
    UCSR0B = (1<<RXEN0)|(1<<TXEN0);
     
    /*Set baud rate */ 
    UBRR0H = (unsigned char)(ubrr>>8); 
    UBRR0L = (unsigned char)ubrr; 
    
    /*Enable receiver and transmitter */ 
    UCSR0B = (1<<RXEN0)|(1<<TXEN0);
   
    /* If the tinyloader isn't loaded yet, we need to fake that we are in it */ 
    uart_putc('?');
    uart_putc('A');
    
    /* Wait for SOT */
    do
        {
        lub_GetPacket(buffer, 10, 1);        
        }while (buffer[2] != LUB_SOT);
    
    /* Sign on message version info */
    buffer[0] = 0x02;
    buffer[1] = 0x00;
        
    /* Send sign-on message */
    lub_SendPacket(LUB_SIGN_ON, buffer, 9);   
    
    /* Now until we get EOT start processing */
    /* Start writing to DF on first page */
    DF_buffernum = 1;
    do
        {
        lub_GetPacketStream(ProcessLubData, FlashData, 9);

        //Requested CRC16
        if (Global_packettype == LUB_CRC1)
            {
            DealWithCRC();
            }

        if (Global_packettype == LUB_SUICIDE)
            {
            CommitSuicide(); 
            }
              
        } while(Global_packettype != LUB_EOT);

        
    /* AVR: I can't deal with the pain of it anymore
       DOC: Please no, you can just jump to the bootloader, there's no
            need to reset
       AVR: It's not worth it.. here! */
    wdt_enable(WDTO_15MS);
    
    /*DOC: Oh no, please reset your WDT, you still have too much
           to live for.
     */
    while(1 == 1)
        {
        continue;
        }
    /*BROTHER: Dibs on his RAM
      DOC: <Runs off-stage right crying>
     */          
        
            
    return 1;
    }

   
/* Thus routine processes each byte as it comes in */     
void ProcessLubData(unsigned char data, unsigned int bytenum)
    {
    Write_flash = 0;
    
    if (bytenum == 0)
        {
        //Don't write this to FLASH
        Write_flash = 0;  

        Global_packettype = data;
        }
        
    if (Global_packettype == LUB_SEND_BOOT)
        {
        DF_page = 0x130;
        CRCStartPage = DF_page; 
        }    
        
    if (Global_packettype == LUB_SEND_SAFEBOOT)
        {
        DF_page = 0x160;
        CRCStartPage = DF_page;
        }    
        
    if (Global_packettype == LUB_SEND_AVR)
        {
        DF_page = 0x100; 
        CRCStartPage = DF_page;
        }
    
    if (Global_packettype == LUB_SEND_FPGA)
        {
        DF_page = 0x400; 
        CRCStartPage = DF_page;
        }   
    else if ((Global_packettype == LUB_264_DATA) && (bytenum == 1))
        {
        //Enable
        Buffer_Write_Enable(DF_buffernum, 0);
        
        //Go!
        DF_SPI_RW(data);   
        }   
    else if ((Global_packettype == LUB_264_DATA) && (bytenum == 264))
        {
        DF_SPI_RW(data);   
        //CS Inactive to end stream mode
        DF_CS_inactive;
        
        //We need to write this to FLASH
        Write_flash = 1;
        }  
    else if (Global_packettype == LUB_264_DATA)
        {
        DF_SPI_RW(data);
        }
      
    }
    
void FlashData(void)
    {
    if (Write_flash == 1)
       {
            
      //Wait for last write to go through
       while(!(Read_DF_status() & 0x80));              
                    
       Buffer_To_Page(DF_buffernum, DF_page);   
                
       //Select next page
        DF_page++;
                    
       //Select other buffer that is current being written
       if (DF_buffernum == 1)
             DF_buffernum = 2;
       else
             DF_buffernum = 1;  
       }

    Write_flash = 0; 
    }
    

void DealWithCRC(void)
    {
    unsigned int intcounter;
    unsigned int pagecounter;
    
    unsigned char buffer[10];

    //Wait for last write to go through
    while(!(Read_DF_status() & 0x80)); 

    Cont_Flash_Read_Enable(CRCStartPage, 0x00);

    crc8_clear();
    intcounter = 0;
    pagecounter = CRCStartPage;

    while (pagecounter != DF_page)
        {
        crc8_addbyte(DF_SPI_RW(0x00));
        intcounter++;
        if (intcounter == 264)
           {
           intcounter = 0;
           pagecounter++;
           }
        }
    
    DF_CS_inactive;

    /* Sign on message version info */
    buffer[0] = crc8_getcrc();
    buffer[1] = 0x00;
        
    /* Send CRC message */
    lub_SendPacket(LUB_CRC1, buffer, 9);       

    return;
    }

/* Commits suicide, aka erases DataFLASH */
void CommitSuicide(void)
    {
    unsigned char block = 0;

    Erase_Block(block);

    /*Kill off the DataFLASH */
    do
        {
        block++;
        Erase_Block(block);
        }while(block != 0xFF);

    }
