/********************************************************************
* PROJECT: LoonBoard Project
*
* COYRIGHT: Colin O'Flynn of NewAE, (C) 2005 <coflynn@newae.com>
*
* LICENSE: This file is released under a license specified in the LICENSE
*          file in the main project directory that this file was in. If
*          the license file is missing, then this file is NOT licensed
*          and all standard copyright laws apply. 
*
* FILE NAME: twi_comms.c                                     
* 
* PURPOSE: Two Wire Interface routine for AVR
* 
* FILE REFERENCES: 
* 
* Name                I/O   Description 
* ----                ---   ----------- 
* none
*
* ABNORMAL TERMINATION CONDITIONS, ERROR AND WARNING MESSAGES: 
* 
* ASSUMPTIONS, CONSTRAINTS, RESTRICTIONS: F_CPU must be system
*    frequency in Hz
* 
* NOTES: 
* 
* REQUIREMENTS/FUNCTIONAL SPECIFICATIONS REFERENCES: 
* 
* DEVELOPMENT HISTORY: 
* 
* Date(DMY)   Author          Release   Description Of Change 
* ----        ------          -------   --------------------- 
* 15-02-2005  Colin O'Flynn    0.1      First working release
*
*******************************************************************/

#include <avr/io.h>
#include <compat/twi.h>

#include "twi_comms.h"
#include "error.h"

#ifndef F_CPU
#error "Error: F_CPU Undefined"
#endif

//TWI prescaller
#define TWPRESCALER 1

/*From datasheet:

TWCLK = SYSCLK/(16 + 2*(TWBR)*(TWPRESCALE))

(sysclk/twclk - 16)/(2*prescale) = twbr

So:

TWBR = (SYSCLK/TWCLK - 16) / (2*TWPRESCALE)
*/

#define TWBRval (uint8_t)(((F_CPU / TWCLK) - 16) / (2*TWPRESCALER))

#define twi_timeout_int() loop_until_bit_is_set(TWCR, TWINT)
                                    
void twi_init(void)
    {
    
    /* Setup prescaler */
    #if TWPRESCALER == 1
      TWSR = 0;
    #elif TWIPRESCALER == 4
      TWSR = 1;
    #elif TWIPRESCALER == 16
      TWSR = 2;
    #elif TWIPRESCALER == 64
      TWSR = 3;
    #endif //TWPRESCALER
        
    TWBR = TWBRval;
    
    /* According to datasheet, if in master mode we need TWBR to be at least
    10 for reliable operation */
    if (TWBR < 10)
        {
        TWBR = 10;
        }
        
    TWCR = 1<<TWEN | 1<<TWEA;
    
    return;
    }
    
void twi_stop(void)
    {
    TWCR = (1<<TWINT) | (1<<TWSTO) | (1<<TWEN);
    return;
    }
	     

ERROR_T twi_send(uint8_t sla, uint8_t * data, unsigned char numbytes)
    {
    unsigned char currbyte;
    uint8_t   status;
    
    //send start condition
    TWCR =(1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
    
    // Wait for TWINT Flag set. This indicates that the START condition has been transmitted
    twi_timeout_int();
    
    switch(TW_STATUS)
        {
        case TW_REP_START:
        case TW_START:
            break;
       
        default:
            return TW_STATUS + TWI_OFFSET;
        }
    
    //Load SLA_W into TWDR Register.        
    TWDR = sla | TW_WRITE;
    TWCR = (1<<TWINT) | (1<<TWEN); 
    
    twi_timeout_int();
    
    status = TW_STATUS;
    switch(status)
        {
        case TW_MT_SLA_ACK:
            break;
       
        default:
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);  
            return status + TWI_OFFSET;
        }
    
    
    currbyte = 0;
    
   
    while(currbyte < numbytes)
        {

        TWDR = data[currbyte];
                       
        TWCR = (1<<TWINT) | (1<<TWEN);
              
        twi_timeout_int();
                
        status = TW_STATUS;
        switch(status)
            {
            case TW_MT_DATA_ACK:
                break;
            default:
                TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);  
                return status + TWI_OFFSET;
            }
        
        currbyte++;
                    
        }    
        
        
    return ERROR_NONE;
    }
    
ERROR_T twi_send_data(uint8_t * data, unsigned char numbytes)
    {
    unsigned char currbyte;
    uint8_t   status;  
    
    currbyte = 0;
    
   
    while(currbyte < numbytes)
        {

        TWDR = data[currbyte];
                       
        TWCR = (1<<TWINT) | (1<<TWEN);
              
        twi_timeout_int();
                
        status = TW_STATUS;
        switch(status)
            {
            case TW_MT_DATA_ACK:
                break;
            default:
                TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);  
                return status + TWI_OFFSET;
            }
        
        currbyte++;
                    
        }    
        
        
    return ERROR_NONE;
    }
    
ERROR_T twi_get(uint8_t sla, uint8_t * data, unsigned char numbytes)
    {
    unsigned char currbyte;
    uint8_t status;
    
    //send start condition
    TWCR =(1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
    
    // Wait for TWINT Flag set. This indicates that the START condition has been transmitted
    twi_timeout_int();
    
    status = TW_STATUS;
    switch(status)
        {
        case TW_REP_START:
        case TW_START:
            break;
       
        default:
            return status + TWI_OFFSET;
        }
    
    //Load SLA_R into TWDR Register.        
    TWDR = sla | TW_READ;
    TWCR = (1<<TWINT) | (1<<TWEN); 
    
    twi_timeout_int();
    
    status = TW_STATUS;
    switch(status)
        {
        case TW_MR_SLA_ACK:
            break;
       
        default:
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);  
            return status + TWI_OFFSET;
        }
        
    currbyte = 0;
    
    while(currbyte < numbytes)
        {
        /*If this is the last byte we NACK it, otherwise ACK it */
        if (currbyte + 1 == numbytes)
            {
            TWCR = (1<<TWINT) | (1<<TWEN);
            }
        else
            {
            TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);
            }
        
        twi_timeout_int();
        
    status = TW_STATUS;
    switch(status)
            {
            case TW_MR_DATA_ACK:
            case TW_MR_DATA_NACK:
                break;
            default:
                TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);  
                return status + TWI_OFFSET;
            }
        
        data[currbyte] = TWDR;
        
        currbyte++;
        }
                    
    return ERROR_NONE;
    }
    
        
 