/********************************************************************
* PROJECT: LoonBoard Project
*
* COYRIGHT: Colin O'Flynn of NewAE, (C) 2005 <coflynn@newae.com>
*
* LICENSE: This file is released under a license specified in the LICENSE
*          file in the main project directory that this file was in. If
*          the license file is missing, then this file is NOT licensed
*          and all standard copyright laws apply. 
*
* FILE NAME: tinyloader.S                                                 
* 
* PURPOSE: This resident bootloader is always present in the AVR.
*          It loads the actual main bootloader into the AVR space,
*          as well as loading the FPGA.
* 
*
* ABNORMAL TERMINATION CONDITIONS, ERROR AND WARNING MESSAGES: 
* 
* ASSUMPTIONS, CONSTRAINTS, RESTRICTIONS: 
* 
* NOTES: 
* 
* REQUIREMENTS/FUNCTIONAL SPECIFICATIONS REFERENCES: 
* Pin Connections:
* PORT   PIN   DIR   USE
* ----   ---   ---   ---
* PORTC   0    OUT   Data Out to FPGA
* PORTD   0    IN    RX from PC (dedicated AVR hardware)
*
* PORTD   1    OUT   TX to PC   (dedicated AVR hardware)
* PORTD   2    OUT   Clock Out to FPGA
* PORTD   7    OUT   to PROGRAM pin on AVR
*
* PORTB   3    OUT   MOSI to DataFLASH   (dedicated AVR hardware)
* PORTB   4    IN    MISO from DataFLASH (dedicated AVR hardware)
* PORTB   5    OUT   SCK to DataFLASH    (dedicated AVR hardware)
* PORTB   7    OUT   CS to DataFLASH
*
* ADC     6    IN    DONE from FPGA
* ADC     7    IN    INIT from FPGA
* 
* DEVELOPMENT HISTORY: 
* 
* Date(DMY)   Author          Release   Description Of Change 
* ----        ------          -------   --------------------- 
* 30-04-2005  Colin O'Flynn   Initial   Initial file written
* 14-05-2005  Colin O'Flynn   0.1       Bug fixes, it works
* 20-05-2005  Colin O'Flynn   0.2       Optimized slightly
* 11-10-2005  Colin O'Flynn   0.3       Changed DataFLASH locations
*
*******************************************************************/

#include <avr/io.h>

#define temp1           R16
#define temp2           R17
#define zero_reg        R20

#define DF_highpagereg  R18
#define DF_lowpagereg   R19
#define DF_datareg      R21

;Registers used for UART
#define UART_UDR        R23

;Registers used for AVR
#define AVR_pagecount   R22
#define looplo          R24
#define loophi          R25
#define spmcrval        R26

#define fpgaload          R28

#define FPGA_INIT   7
#define FPGA_DONE   6

#define PORT_IO(x)     _SFR_IO_ADDR(x)
#define PORT_MEM(x)    _SFR_MEM_ADDR(x)

;Get pagesize in bytes
#ifndef SPM_PAGESIZE
    #error SPM_PAGESIZE NOT DEFINED
#endif

#define PAGESIZEB SPM_PAGESIZE

.section .text
.global main

main:

    ;There is a chance some sort of WDT was on, we need to
    ;disable that or we are screwed
    
    ; Reset Watchdog Timer
    wdr
    ; Clear WDRF in MCUSR
    in temp1, PORT_IO(MCUSR)
    andi temp1, (0xff & (0<<WDRF))
    out PORT_IO(MCUSR), temp1
    ; Write logical one to WDCE and WDE
    ; Keep old prescaler setting to prevent unintentional time-out
    lds temp1, PORT_MEM(WDTCSR)
    ori temp1, (1<<WDCE) | (1<<WDE)
    sts PORT_MEM(WDTCSR), temp1
    ; Turn off WDT
    ldi temp1, (0<<WDE)
    sts PORT_MEM(WDTCSR), temp1

    ldi temp1, hi8(RAMEND)   ;Stack-pointer needs setup
    out PORT_IO(SPH), temp1
    ldi temp1, lo8(RAMEND)
    out PORT_IO(SPL), temp1

    clr zero_reg            ;Zero-reg is always cleared
    
    ;Get rid of interrupts
    cli
    
    ;Enable receiver and transmitter
    ldi temp1, (1<<RXEN0) | (1<<TXEN0)
    sts PORT_MEM(UCSR0B), temp1
    
    ;Get that baudrate right    
    ;XTAL       BAUD     UBRR
    ;6.75MHz   57600     6
    ;7.37MHz   115200    3 
    ;7.37MHz   57600     7
    ldi temp1, 0x03
    sts PORT_MEM(UBRR0L), temp1
    sts PORT_MEM(UBRR0H), zero_reg
 
    out PORT_IO(PORTD), zero_reg ;Program is low behind-the-scenes
    
    ;Set SS, MOSI, SCK, and the real CS connected to our dataflash's outputs
    ldi temp1, 1<<7 | 1<<2 | 1<<0; ;Also red LED is on
    out PORT_IO(PORTB), temp1    
    ldi temp1, 1<<2 | 1<<7 | 1<<3 | 1<<5 | 1<<1 | 1<<0;
    out PORT_IO(DDRB), temp1

    ;SPI Interface active
    ldi temp1, 1<<SPE | 1<<MSTR | 1<<CPHA | 1<<CPOL
    out PORT_IO(SPCR), temp1
 
    ;Have to wait for DataFLASH to come up
    rcall delay2
       
    ;Read OSCCAL, address 0 on page 1
    ldi DF_highpagereg, 0x00
    ldi DF_lowpagereg, 0x01    
    rcall DF_Cont_Read 
    rcall DF_Send_Rec  
    sts PORT_MEM(OSCCAL), DF_datareg
    sbi PORT_IO(PORTB), 7   
     
    ;Make sure oscillator is stable
    rcall delay2 
   
    ;UART Debugging routine
   ;aoe:
   ; lds temp1, UCSR0A
   ; sbrs temp1, RXC0
   ; rjmp aoe
   ; 
   ; lds UART_UDR, UDR0
   ; rcall UART_Transmit
   ; rjmp aoe
    
   ;Don't load FPGA by default
    clr fpgaload
    
  ;***Startup - Try communicating with computer
    ;Send sync byte of ?
    ldi UART_UDR, '?'
    rcall UART_Transmit
 
    ;Check for a return character, wait a bit first
    rcall delay2
    
    lds UART_UDR, PORT_MEM(UDR0)

    ;a b means load bootloader code
    cpi UART_UDR, 'b'
    breq STARTUP_BOOTLOADER
    
    ;a s means safemode bootloader code
    cpi UART_UDR, 's'
    breq STARTUP_SAFEMODE
    
    ;a l means load the normal AVR code
    cpi UART_UDR, 'l'
    breq STARTUP_NORMAL
    
    ;d is for debugging
    ; cpi UART_UDR, 'd'
    ; breq STARTUP_DEBUG
    
    ;Otherwise we are starting up in normal mode, jump to FPGA loading
    rjmp FPGA_Load
    
    
STARTUP_BOOTLOADER:
    ;Setup location of AVR File (0x0130)
    ldi DF_highpagereg, 0x01
    ldi DF_lowpagereg, 0x30
    rjmp AVR_Load
    
STARTUP_SAFEMODE:
    ;Setup location of AVR File (0x0160)
    ldi DF_highpagereg, 0x01
    ldi DF_lowpagereg, 0x60   
    rjmp AVR_Load
   
;STARTUP_DEBUG:
    ;rjmp Stuck
    
STARTUP_NORMAL:
    ;Setup location of AVR File (0x0100)
    ldi DF_highpagereg, 0x01
    ldi DF_lowpagereg, 0x00 
    ldi fpgaload, 0xFF
 
;***********************************AVR Loading Section****************************************
AVR_Load:
    ;How many pages to write
    ;=128 - Pages used for this bootloader
    ;32 pages in max bootloader section 
    ;One page = 32 words / 64 Bytes
    ldi AVR_pagecount, ((FLASHEND + 1) / PAGESIZEB) - ((FLASHEND - BOOTSTART + 1) / PAGESIZEB)
        
    ;Load Z Pointer with 0
    clr ZL
    clr ZH
    
   ;Enable read from DF
    rcall DF_Cont_Read
    
    ;Now until we have written every page we go 
  Page_Write_Loop:
    ;Write the current page
    rcall Write_page
    ;Select next page, starts at bit 6 for page
    adiw ZL, 63
    adiw ZL, 1
  
    ;Count pages read
    dec AVR_pagecount
    
    BRNE Page_Write_Loop
    
    ;Disable continous read
    sbi PORT_IO(PORTB), 7

    ;If this was a routine AVR load we also do want to load the FPGA
    sbrs fpgaload, 1 
    
    rjmp Done
 
 
;***********************************FPGA Loading Section***************************************        
;Load the FPGA    
FPGA_Load:     

    ;In order to cheat out more I/O pins, the dedicated ADC channels are used as inputs
    ; ADC with AVCC as ref, Left-Adjusted, FPGA_INIT pin as first pin
    ldi temp1, (1<<REFS0) | 1<<ADLAR | FPGA_INIT 
    sts PORT_MEM(ADMUX), temp1
    
    ldi temp1, (1<<ADEN) | (1<<ADSC) | (1<<ADATE) | (1<<ADPS2) | (1<<ADPS0);
    sts PORT_MEM(ADCSRB), zero_reg
    
    ;ADC enabled, start first conversion, free-running, divide clock by 32
    sts PORT_MEM(ADCSRA), temp1 

    ;First step of programming algorithm for FPGA - pulse /PROGRAM low
    sbi PORT_IO(DDRD), 7
    nop
    nop
    nop
    cbi PORT_IO(DDRD), 7
    
    ;Wait for INIT to go high   
  waiting_for_init:
    lds temp1, PORT_MEM(ADCH)
    sbrs temp1, 7 ;If bit is set, INIT is high
    rjmp waiting_for_init
    
    ;Clear CCLK
    cbi PORT_IO(PORTC), 0
    
    ;Wait about 100 uS
    rcall delay  
  
    ;Set DO and CCLK as outputs
    sbi PORT_IO(DDRC), 0
    sbi PORT_IO(DDRD), 2
    
    ;Set ADC to read DONE
    ldi temp1, FPGA_DONE | 1<<REFS0 | 1<<ADLAR
    sts PORT_MEM(ADMUX), temp1
  
    ;Setup location of FPGA File (0x0400)
    ldi DF_highpagereg, 0x04
    ldi DF_lowpagereg, 0x00
    
    ;Enable read from DF
    rcall DF_Cont_Read
  
    ;First three bytes are 24-bit size of bitstream, to be stored in specified values
    rcall DF_Send_Rec
    rcall DF_Send_Rec
    rcall DF_Send_Rec
      
    ;Now this part is easy - pretty much just grab a bit from the
    ;dataflash and put it into the FPGA, since everything is done serially 
    ;
    ;In this case we do 8 bits at a time so we can use the hardware SPI interface. It makes using something
    ;beside DataFLASH easier. It works out as the same code size and about the same speed as doing it software
    ;grab a bit and throw it out the other pin.
    ;                
    ;You could physically connect the CCLK line on the FPGA to the SCK line on the DataFLASH, and
    ;the MISO pin on the DataFLASH and the DI pin on the FPGA if you wanted, which would be faster and
    ;smaller
    ;
    ;Done pin goes high if config OK, or INIT pin goes low if init failed
    ;
  waiting_for_done:
    rcall DF_Send_Rec
    rcall FPGA_Send_Byte
       
    lds temp1, PORT_MEM(ADCH)
    sbrs temp1, 7
    rjmp waiting_for_done
   
    ;Disable continous read
    sbi PORT_IO(PORTB), 7
    
;******** Cleaning Up***********************************************************
Done:         

    ;Set CS as input, external resistors take over
    cbi PORT_IO(DDRB), 7
    
    ;Disable DF registers
    out PORT_IO(SPCR), zero_reg

    ;Green LED On, Red Off
    sbi PORT_IO(PORTB), 1
    cbi PORT_IO(PORTB), 0
  
;Jump home to main application
Jump_home:
    clr ZL
    clr ZH
    ijmp  
 
;Stuck:
    ;Green LED On
 ;   sbi PORT_IO(PORTB), 1
 ;   rjmp Stuck
       
;************************ DataFLASH Subroutines ********************************   
;Pulses CS Pin
DF_Pulse_CS:
    sbi PORT_IO(PORTB), 7
    nop
    cbi PORT_IO(PORTB), 7    
    reti
    
;Waits for transmission from SPI - from Datasheet
DF_Wait_Transmit: ; Wait for transmission complete 
    in temp1, PORT_IO(SPSR)
    sbrs temp1,SPIF 
    rjmp DF_Wait_Transmit 
    reti

DF_Send_Rec:
    out PORT_IO(SPDR), DF_datareg  ;put data out
    rcall DF_Wait_Transmit      ;Wait for trans complete
    in  DF_datareg, PORT_IO(SPDR)  ;Read data in
    reti                        ;Return
    
DF_Cont_Read:
    rcall DF_Pulse_CS
    ldi DF_datareg, 0x68    ;continous read
    rcall DF_Send_Rec
    
    ;Need to send address bytes: 0 0 0 0 A10 A9 A8 A7 then A6 A5 A4 A3 A2 A1 0
    mov temp1, DF_highpagereg
    mov temp2, DF_lowpagereg
    lsl temp2 ;Lower is shifted once, upper bit into carry
    ROL temp1  ;Upper is rolled once, carry goes into lower bit
    mov DF_datareg, temp1
    rcall DF_Send_Rec
    mov DF_datareg, temp2
    rcall DF_Send_Rec
    
    ;Send 5 blank bytes
    mov DF_datareg, zero_reg
    rcall DF_Send_Rec
    rcall DF_Send_Rec
    rcall DF_Send_Rec
    rcall DF_Send_Rec
    rcall DF_Send_Rec
    
    reti

;********************** Subroutines for AVR Loading ***************************
;
;Write one page from DataFLASH to AVR    
;
;DataFLASH must have already been enabled in constant
;read mode at the proper address
;
;Z-Pointer must point to AVR FLASH location to write to
;
  Write_page:
    ; Page Erase
    ldi spmcrval, (1<<PGERS) | (1<<SELFPRGEN)
    rcall Do_spm
    ; re-enable the RWW section
    ldi spmcrval, (1<<RWWSRE) | (1<<SELFPRGEN)
    rcall Do_spm
    ; transfer data from DF to Flash page buffer
    ldi looplo, lo8(PAGESIZEB) ;init loop variable
    ldi loophi, hi8(PAGESIZEB) ;not required for PAGESIZEB<=256
  Wrloop:
    ;Get data from DataFLASH for program
    rcall DF_Send_Rec
    mov R0, DF_datareg
    rcall DF_Send_Rec
    mov R1, DF_datareg
        
    ldi spmcrval, (1<<SELFPRGEN)
    rcall Do_spm
    adiw ZL, 2
    sbiw looplo, 2 ;use subi for PAGESIZEB<=256
    brne Wrloop
    ; execute Page Write
    subi ZL, lo8(PAGESIZEB) ;restore pointer
    sbci ZH, hi8(PAGESIZEB) ;not required for PAGESIZEB<=256
    ldi spmcrval, (1<<PGWRT) | (1<<SELFPRGEN)
    rcall Do_spm
  Return:
    in temp1, PORT_IO(SPMCSR)
    sbrs temp1, RWWSB ; If RWWSB is set, the RWW section is not ready yet
    ret
    ; re-enable the RWW section
    ldi spmcrval, (1<<RWWSRE) | (1<<SELFPRGEN)
    rcall Do_spm
    rjmp Return
    
  Do_spm:
    ; check for previous SPM complete
    ; No EEPROM Bootloading done here so don't worry
    ; about the EEPROM
  Wait_spm:
    in temp1, PORT_IO(SPMCSR)
    sbrc temp1, SELFPRGEN
    rjmp Wait_spm
    out PORT_IO(SPMCSR), spmcrval
    spm
    ret  
    
;****************** Subroutines for FPGA Loading ***********************************    
   

;Use the CLK and DO pins to send a databyte to FPGA
FPGA_Send_Byte:
    ldi temp1, 0x08
    
 FPGA_Send_Byte_loop:
    cbi PORT_IO(PORTD), 2 ;Make sure clock is clear
    
    lsl DF_datareg ;Shift out MSb to Carry Flag
    
    cbi PORT_IO(PORTC), 0;Clear data bit
    
    brcc FPGA_Send_Byte_sendlow ;Check if carry is low
    
    sbi PORT_IO(PORTC), 0 ;Nope - set that bit high
    
 FPGA_Send_Byte_sendlow:
    
    sbi PORT_IO(PORTD), 2; Clock is high
    
    dec temp1
    brbc    1, FPGA_Send_Byte_loop ;Check to see if this is bit 0
    
    reti
    
;***************** Subroutines for UART *****************************************

UART_Transmit:
    ; Wait for empty transmit buffer
    lds temp1, PORT_MEM(UCSR0A)
    sbrs temp1,UDRE0
    rjmp UART_Transmit
    ; Put data (r16) into buffer, sends the data
    sts PORT_MEM(UDR0), UART_UDR
    ret
       
;*************** Delay's ********************************************************
    
;Need a proper delay in FPGA routine of about 100 uS
;Delays 4 cycles per loop, does 256 loops = 1024 cycle delay
delay:
    nop
    dec zero_reg    ;Dec zero_reg
    brbc    1,delay ;If zero flag is clear, we branch
    reti
        
;Just a longer delay
delay2: 
    clr temp1
    delay_start:
    rcall delay
    dec  temp1    ;Dec temp1
    brbc SREG_Z,delay_start ;If zero flag is clear, we branch
    ret
    
    
    
    
    
