/*
 * lubloader - NewAE LoonBoard Unified Bootloader program
 * Copyright (C) 2005  Colin O'Flynn <coflynn@newae.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/*
 * Note: code is based heavily on source code to avrdude, which is
 * Copyright (C) 2000-2004 Brian S. Dean <bsd@bsdhome.com>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "fileio.h"
#include "lub.h"
#include "serial_lowlevel.h"

void usage(void);
int get264Bytes(unsigned char * data, unsigned char * file, unsigned long current_byte, unsigned long size);
int sendLUBBuffer(unsigned char datatype, unsigned char * file_buffer, unsigned long size_of_file);
unsigned char getCRC(void);


const char progname[] = "lubloader";
int     verbose;     /* verbose output */
unsigned char max_tries;    /* Maximum allowed retries */
    
int main (int argc, char * argv [])
    {
    char    data[270];    /* Used for packet buffer */    
    char    ch;          /* options flag */
    char  * e;           /* for strtol() error checking */
    char  * port;        /* device port (/dev/xxx) */
    char  * file_fpga;   /* filename of FPGA file */
    char  * file_avr;    /* filename of AVR file */
    char  * file_lub;    /* filename of AVR LUB file */
    char  * file_safelub;    /* filename of AVR LUB file */
    long    baudrate;    /* baud rate */
    
    char    bootloader;  /* Set to 1 for safemode bootloader*/
                         /* Set to 0 for normal bootloader */  
                         /* Set to 3 for debug */ 

    char    erase;       /* Set to 1 to erase DataFLASH */
      
    FILE *  fpga_fptr;  /* Pointer to FPGA file */
    FILE *  avr_fptr;   /* Pointer to AVR file */
    FILE *  lub_fptr;   /* Pointer to AVR LUB file */
    FILE *  safelub_fptr;   /* Pointer to AVR Safemode LUB file */
    
    long size_of_file;     /* Size of file read */
    long size_of_file2;     /* Size of file read */
    unsigned char    file_buffer_real[1000000];
    unsigned char *  file_buffer;   /* 1MB file buffer */
    
    unsigned char   local_crc;

    const char    empty_string[] = {'\0'};
    
    port        = empty_string;
    file_fpga   = empty_string;
    file_avr    = empty_string;    
    file_lub    = empty_string;
    file_safelub= empty_string;
    port        = empty_string;
    
    fpga_fptr       = !NULL;
    avr_fptr        = !NULL;
    lub_fptr        = !NULL;
    safelub_fptr    = !NULL;
    
    baudrate    = 115200;
    verbose     = 0;
    bootloader  = 0;
    max_tries   = 5;
    
    if (argc == 1)
       {
       usage();
       exit(1);
       }
    
    
    /*
     * process command line arguments
     */
    while ((ch = getopt(argc,argv,"hdevsq:w:b:P:f:a:")) != -1) 
        {

        switch (ch) 
        {
        case 'b': /* override default baud rate */
            baudrate = strtol(optarg, &e, 0);
            if ((e == optarg) || (*e != 0)) {
            fprintf(stderr, "%s: invalid baud rate specified '%s'\n",
                    progname, optarg);
            exit(1);
            }
            break;
        case 'd':
            bootloader = 3;
            break;
        
        case 's':
            bootloader = 1;
            break;
    
        case 'P':
            port = optarg;
            break;
        
        case 'e':
            erase = 1;
            break;

        case 'f':
            file_fpga = optarg;
            break;
            
        case 'a':
            file_avr = optarg;
            break;
            
        case 'q':
            file_lub = optarg;
            break;
            
        case 'w':
            file_safelub = optarg;
            break;
    
        case 'v':
            verbose++;
            break;
    
        case 'h': /* help */
            usage();
            exit(0);
            break;
    
        default:
            fprintf(stderr, "%s: invalid option -%c\n\n", progname, ch);
            usage();
            exit(1);
            break;
        }

        }
            
    if (port == empty_string)
        {
        fprintf(stderr, "%s: ERROR: You must specify a serial port with -P!\n", progname);
        exit(1);
        }
        
    if (verbose > 0)
        {
        fprintf(stderr, "%s: Using serial port: %s at baud of %d\n",progname, port, baudrate);
        fprintf(stderr, "%s: Using FPGA filename as: %s\n",progname, file_fpga);
        fprintf(stderr, "%s: Using AVR filename as: %s\n", progname, file_avr);
        fprintf(stderr, "%s: Using BOOTLOADER filename as: %s\n", progname, file_lub);
        fprintf(stderr, "%s: Using SAFEMODE BOOTLOADER filename as: %s\n", progname, file_safelub);
        }
        
    /* Sanity checking. Cannot re-write safemode bootloader if we are going to use safemode to enter the
       bootloader, doing so would be suicide.
       
       Also do not write both safemode and normal mode bootloader at once. Again, suicide */
    if ((file_lub != empty_string) && (file_safelub != empty_string))
        {
        fprintf(stderr, "%s: SAFTEY OVERRIDE: You are attempting to write both the safemode and\n"
                        "%s:                  normal mode bootloaders at once. This is not allowed\n"
                        "%s:                  since doing so could result in a dead target if both\n"
                        "%s:                  files were corrupt.\n", progname, progname, progname, progname);
        exit(1);
        }
        
   if ((file_safelub != empty_string) && (bootloader == 1))
        {
        fprintf(stderr, "%s: SAFTEY OVERRIDE: You are attempting to write the safemode bootloader\n"
                        "%s:                  while running from the safemode bootloader. If the file\n"
                        "%s:                  was corrupt you may have no working bootloaders on the\n"
                        "%s:                  target. Write the normal bootloader and program the safemode\n"
                        "%s:                  bootloader from there.\n", progname, progname, progname, progname, progname);
        exit(1);        
        
        }
        
        
    /* Figure out which files we should be using, open them */
    if (file_fpga != empty_string)
        {
        fpga_fptr = fopen(file_fpga, "rb");
        }
    if (file_avr != empty_string)
        {
        avr_fptr = fopen(file_avr, "rb");
        }
    if (file_lub != empty_string)
        {
        lub_fptr = fopen(file_lub, "rb");
        }
    if (file_safelub != empty_string)
        {
        safelub_fptr = fopen(file_safelub, "rb");
        }
        
               
    /* Check we did actually open the files */
    if (fpga_fptr == NULL)
        {
        fprintf(stderr, "%s: Unable to open file %s\n", progname, file_fpga); 
        exit(1);
        }
    if (avr_fptr == NULL)
        {
        fprintf(stderr, "%s: Unable to open file %s\n", progname, file_avr); 
        exit(1);
        }
     if (lub_fptr == NULL)
        {
        fprintf(stderr, "%s: Unable to open file %s\n", progname, file_lub); 
        exit(1);
        }
     if (safelub_fptr == NULL)
        {
        fprintf(stderr, "%s: Unable to open file %s\n", progname, file_safelub); 
        exit(1);
        }
  
      
    /* Open the serial port */
    serial_init(port, baudrate);
            
        
    /* Communicate with LUB */
    fprintf(stderr, "\n%s: Attempting to communicate with Loonboard Unified Bootloader\n", progname);
    fprintf(stderr, "%s: Power cycle device now to initiate LUB\n\n", progname);
    
    /* The LightLUB can understand three commands:     's' for safe mode bootloader
                                                       'b' for normal bootloader
                                                       'l' for normal startup 
    */
    if (lub_EnterLUB(bootloader) != 0)
        {
        fprintf(stderr, "%s: Unable to enter Loonboard Unified Bootloader\n", progname);
        exit(1);
        }
    
    /* Send Start Of Transmission */
    if (verbose > 0)
        fprintf(stderr, "%s: Sending SOT...\n", progname);
    
    if (lub_SendPacket(LUB_SOT, NULL, max_tries) != 0)
        {
        fprintf(stderr, "%s: Comm error\n", progname);
        exit(1);
        }
    
    if (lub_GetPacket(data, 10, max_tries) != 0) 
        {
        fprintf(stderr, "%s: Comm error\n", progname);
        exit(1);
        } 
    
    if (data[2] != LUB_SIGN_ON)
        {
        fprintf(stderr, "%s: Received %x, Expecting %x\n", progname, data[2], LUB_SIGN_ON);
        exit(1);
        }
        
    if (verbose > 0)
        {
        fprintf(stderr, "%s: LUB Minor vers: %d\n", progname, data[3]);
        fprintf(stderr, "%s: LUB Major vers: %d\n", progname, data[4]);
        }

  if (erase == 1)
     {
     fprintf(stderr, "%s: ERASE Of DataFLASH in progress.\n", progname);
     if (lub_SendPacket(LUB_SUICIDE, NULL, max_tries) != 0)
        {
        fprintf(stderr, "%s: Comm error\n", progname);
        exit(1);
        }
     fprintf(stderr, "%s: DataFLASH ERASED.\n", progname);
     }


  /************* SEND FPGA FILE *****************/
 
  if (file_fpga != empty_string)
    {
  
    /* Open FPGA File, assign to buffer */
    file_buffer = file_buffer_real;
    size_of_file = fread(file_buffer, 1, 1000000, fpga_fptr);
    
    if (size_of_file > 1000000)
        {
        fprintf(stderr, "%s: Internal limit, file for FPGA must be smaller than 1MB\n", progname);
        exit(1);
        }
        
    /* Okay... so we need to process the header somewhat, find the point with 0x65 0x00 */
    while (((*file_buffer != 0x65) || (*(file_buffer + 1) != 00)) && (size_of_file != 0))
        {
        file_buffer++;
        size_of_file--;
        continue;
        }
    

    if (size_of_file == 0)
        {
        fprintf(stderr, "%s: Unable to read file. Did you specify the proper option?\n", progname);
        exit(1);
        }

    file_buffer += 2;
            
    /* The next three bytes will be the size of the file */
    size_of_file2 = *(file_buffer) * (65536);
    size_of_file2 += *(file_buffer+1) * (256);
    size_of_file2 += *(file_buffer+2) * (1);  
    
    /* Check that the sizes are the same */
    size_of_file -= 5;
    
    if (size_of_file2 != size_of_file)
        {
        fprintf(stderr, "%s: File size mismatch between reported and actual.\n", progname);
        fprintf(stderr, "%s: Reported: %d  Actual: %d.\n", progname, size_of_file2, size_of_file);
        exit(1);
        }
        
    if (verbose > 0)
        {
        fprintf(stderr, "%s: FPGA Bitstream has size of %d bytes\n", progname, size_of_file);
        }
        
    /* We send three bytes of size info first though... */
    size_of_file += 3;
    
    //Clear CRC
    crc8_clear1();

    fprintf(stderr, "%s: Sending FPGA bit file\n", progname);         
    sendLUBBuffer(LUB_SEND_FPGA, file_buffer, size_of_file);   

    //Save local checksum
    local_crc = crc8_getcrc1();

     //Check CRC
     if (verbose > 0)
        {
        fprintf(stderr, "%s: Local checksum: %X\n", progname, local_crc);
        }   

    if (getCRC() == local_crc)
        {
        fprintf(stderr, "%s: Program verify OK\n", progname);
        }
    else
        {
        fprintf(stderr, "%s: Program verify FAILED\n", progname);
        }
        

    }
    
  /************************* Send LUB File **********************/
 
  if (file_lub != empty_string)
    {
    fprintf(stderr, "%s: Sending Bootloader file\n", progname);
    
    
    /* Convert AVR ihex to bin */
    
    file_buffer = file_buffer_real;
    
    size_of_file = (long)ihex2b(file_lub, lub_fptr, file_buffer, 1000000);
    
    if (size_of_file < 0)
        {
        fprintf(stderr, "%s: AVR LUB hex file error\n", progname);
        exit(1);
        }
  
    if (size_of_file > 7680)
        {
        fprintf(stderr, "%s: WARNING: AVR binary may be too big. You will not be told again.\n", progname);
        }
        
         
    if (verbose > 0)
        {
        fprintf(stderr, "%s: AVR LUB binary has size of %d bytes\n", progname, size_of_file);
        }
        
    /* We pad the file with 0xFF */
    while (size_of_file != 7680)
        {
        file_buffer[size_of_file] = 0xFF;
        size_of_file++;
        }
    
    //Clear CRC
    crc8_clear1();
    
    //Send data
    sendLUBBuffer(LUB_SEND_BOOT, file_buffer, size_of_file);

    //Save local checksum
    local_crc = crc8_getcrc1();

     //Check CRC
     if (verbose > 0)
        {
        fprintf(stderr, "%s: Local checksum: %X\n", progname, local_crc);
        }   

    if (getCRC() == local_crc)
        {
        fprintf(stderr, "%s: Program verify OK\n", progname);
        }
    else
        {
        fprintf(stderr, "%s: Program verify FAILED\n", progname);
        }
        
  }
    
    
  /************************* Send Safemode LUB File **********************/
 
  if (file_safelub != empty_string)
    {
    fprintf(stderr, "%s: Sending Safemode Bootloader file\n", progname);
    
    
    /* Convert AVR ihex to bin */
    
    file_buffer = file_buffer_real;
    
    size_of_file = (long)ihex2b(file_safelub, safelub_fptr, file_buffer, 1000000);
    
    if (size_of_file < 0)
        {
        fprintf(stderr, "%s: AVR Safemode LUB hex file error\n", progname);
        exit(1);
        }
  
    if (size_of_file > 7680)
        {
        fprintf(stderr, "%s: WARNING: AVR Safemode binary may be too big. You will not be told again.\n", progname);
        }
        
         
    if (verbose > 0)
        {
        fprintf(stderr, "%s: AVR Safemode LUB binary has size of %d bytes\n", progname, size_of_file);
        }
        
    /* We pad the file with 0xFF */
    while (size_of_file != 7680)
        {
        file_buffer[size_of_file] = 0xFF;
        size_of_file++;
        }
        
    //Clear CRC
    crc8_clear1();

    sendLUBBuffer(LUB_SEND_SAFEBOOT, file_buffer, size_of_file);

    //Save local checksum
    local_crc = crc8_getcrc1();

     //Check CRC
     if (verbose > 0)
        {
        fprintf(stderr, "%s: Local checksum: %X\n", progname, local_crc);
        }   

    if (getCRC() == local_crc)
        {
        fprintf(stderr, "%s: Program verify OK\n", progname);
        }
    else
        {
        fprintf(stderr, "%s: Program verify FAILED\n", progname);
        }
        
    }
  
   
  /************************* Send AVR File **********************/
    
  if (file_avr != empty_string)
    {
    
    /* Convert AVR ihex to bin */
    
    file_buffer = file_buffer_real;
    
    size_of_file = (long)ihex2b(file_avr, avr_fptr, file_buffer, 1000000);
    
    if (size_of_file < 0)
        {
        fprintf(stderr, "%s: AVR hex file error\n", progname);
        exit(1);
        }

    if (size_of_file > 7680)
        {
        fprintf(stderr, "%s: WARNING: AVR binary may be too big. You will not be told again.\n", progname);
        }
        
          
         
    if (verbose > 0)
        {
        fprintf(stderr, "%s: AVR binary has size of %d bytes\n", progname, size_of_file);
        }
        
    /* We send three bytes of size info first though... */
    //size_of_file += 3;
    
    /* We pad the file with 0xFF */
    while (size_of_file != 7680)
        {
        file_buffer[size_of_file] = 0xFF;
        size_of_file++;
        }
    
    fprintf(stderr, "%s: Sending AVR file\n", progname);  

    //Clear CRC
    crc8_clear1();
  
    sendLUBBuffer(LUB_SEND_AVR, file_buffer, size_of_file);

    //Save local checksum
    local_crc = crc8_getcrc1();

     //Check CRC
     if (verbose > 0)
        {
        fprintf(stderr, "%s: Local checksum: %X\n", progname, local_crc);
        }   

    if (getCRC() == local_crc)
        {
        fprintf(stderr, "%s: Program verify OK\n", progname);
        }
    else
        {
        fprintf(stderr, "%s: Program verify FAILED\n", progname);
        }
        
    }
      
  /* Send End of Transmission (EOT) */
    
            
  if (verbose > 0)
      {
      fprintf(stderr, "%s: Sending EOT Packet\n", progname);
      }
    
  if (lub_SendPacket(LUB_EOT, NULL, max_tries) != 0)
      {
      fprintf(stderr, "%s: Comm error.\n", progname);
      exit(1);
      }
      
  /* Reset the loader to force it to load the new AVR code */ 
  
  fprintf(stderr, "%s: Reseting AVR and forcing load of new code\n", progname);
  
  //No longer true
  //fprintf(stderr, "%s: FPGA code will still not be loaded until hard reset\n", progname);
    
  while (serial_getc() != '?')
      {
      continue;
      }
      
  serial_putc('l');
     
  exit(0);    

  }
    
/*
 * Send a buffer
 */
int sendLUBBuffer(unsigned char datatype, unsigned char * file_buffer, unsigned long size_of_file)
    {
    unsigned char data[264];
    unsigned char returnvalue;
    unsigned int progress;
    unsigned long current_byte;
    
    /* We are now in the main LUB, so first send the fact we want to store a new AVR file */
     if (verbose > 0)
            {
            fprintf(stderr, "%s: Sending request to store file.\n", progname);
            }
        //send request
       if (lub_SendPacket(datatype, NULL, max_tries) != 0)
        {
        fprintf(stderr, "%s: LUB not responding\n", progname);
        return -1;
        }
    
    progress = 0;
    current_byte = 0;
    report_progress (progress, size_of_file / 264, "Writing");
    /* Send a 264 byte data packet until we are done */
    do
        {
        progress++;
        returnvalue = get264Bytes(data, file_buffer, current_byte, size_of_file);
        report_progress (progress, size_of_file / 264, NULL);

        if (verbose > 1)
            {
            fprintf(stderr, "%s: Sending 264 byte data packet\n", progname);
            }
            
        //Send data
        if (lub_SendPacket(LUB_264_DATA, data, max_tries) != 0)
            {
            fprintf(stderr, "%s: Data comm error.\n", progname);
            return -1;
            }
                   
            
        current_byte += 264;
        }while(returnvalue == 0);
   
    if (verbose > 0)
        {
        fprintf(stderr, "%s: File sent\n", progname);
        }
    
    return 0;
    }        
    
    
         
/*
 * Get 264 Bytes
 */  
int get264Bytes(unsigned char * data, unsigned char * file, unsigned long current_byte, unsigned long size)
    {    
    unsigned int current_count = 0;
    
    //Do 264 bytes
    while(current_count != 264)
        {
        //Note: since current_byte starts at 0, and size starts at 1,
        //every single byte will be read and there is no problem with
        //using a < here
        if (current_byte < size)
            {
            data[current_count] = file[current_byte];
            }
        else
           {
           //Fill unused with 0xFF
           data[current_count] = 0xFF;
           }
        
        current_count++;
        current_byte++;
        }  
    
    if (current_byte < size)        
        return 0;
    else
        return -1;
             
   }
         
/*
 * usage message
 */
void usage(void)
{
    fprintf(stderr,
    "Example Usage:"
    "  %s -P /dev/ttyS0 -f hardware_interface.bit -a main.hex -q lub.hex -v -b 57600\n"
    "  %s -P com1 -a main.hex -b 57600\n"
    "  %s -P /dev/ttyS0 -b 57600\n"
    "    Note: This last example will force the bootloader to load from DataFLASH to AVR\n"
    "          the normal run-time code\n"
    " "
    "Usage: %s [options]\n"
    "Options:\n"
    "  -P <port>                  Serial port device.\n"
    "  -s                         Use safemode bootloader.\n"
    "  -f <filename>              .bit file to download to FPGA\n"
    "  -a <filename>              .hex file to download to AVR\n"
    "  -q <filename>              .hex file for new bootloader\n"
    "  -w <filename>              .hex file for new safemode bootloader\n"
    "  -v                         Verbose output, use -vv for more\n"
    "  -b <baudrate>              Use baudrate specified\n"
    "  -e                         ERASE the DataFLASH: ONLY FOR DEBUGGING\n"
    "                               WILL RESULT IN A DEAD SYSTEM\n"
    "  -d                         Enters debug mode in bootloader\n"
    "  -h                         Display this usage.\n"
    "\nLoonBoard Unified Bootloader (LUB) project: http://www.newae.com\n"
    ,progname, progname, progname, progname);
}

unsigned char getCRC(void)
    {
    unsigned char    data[10];    /* Used for packet buffer */    
    
    /* Check file */
    fprintf(stderr, "%s: Instructing lub to calculate CRC\n", progname);
    if (lub_SendPacket(LUB_CRC1, data, max_tries) != 0)
        {
        fprintf(stderr, "%s: Comm error\n", progname);
        exit(1);
        }
    
    if (lub_GetPacket(data, 10, max_tries) != 0) 
        {
        fprintf(stderr, "%s: Comm error\n", progname);
        exit(1);
        } 
    
    if (verbose > 0)
        {
        fprintf(stderr, "%s: Remote checksum: %X\n", progname, data[3]);
        } 

    return (data[3]);
    }


