/*
 * avrdude - A Downloader/Uploader for AVR device programmers
 * Copyright (C) 2000-2004  Brian S. Dean <bsd@bsdhome.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 /*
  *Note: This code is used in the lubloader program here, slightly modified
  */

#include <stdio.h>
#include <time.h>
#include <sys/time.h>

#include "report_progress.h"

/* mingw doesn't support gettimeofday, make our own */

#ifdef GETTIMEOFDAY
/* These are winbase.h definitions */

   typedef struct _FILETIME {
    unsigned long dwLowDateTime;
    unsigned long dwHighDateTime;
    } FILETIME;

void __stdcall GetSystemTimeAsFileTime(FILETIME*);

void gettimeofday(struct timeval* p, void* tz /* IGNORED */);
   
void gettimeofday(struct timeval* p, void* tz /* IGNORED */)
    {
    union 
        {
        long long ns100; /*time since 1 Jan 1601 in 100ns units */
        FILETIME ft;
        } _now;


    GetSystemTimeAsFileTime( &(_now.ft) );
    p->tv_usec=(long)((_now.ns100 / 10LL) % 1000000LL );
    p->tv_sec= (long)((_now.ns100-(116444736000000000LL))/10000000LL);
    return;
    }
#endif /* GETTIMEOFDAY */

static void update_progress_tty (int percent, double etime, char *hdr);
static void update_progress_no_tty (int percent, double etime, char *hdr);

typedef void (*FP_UpdateProgress)(int percent, double etime, char *hdr);
  
static FP_UpdateProgress update_progress;

/* Report the progress of a read or write operation from/to the device.

   The first call of report_progress() should look like this (for a write op):

     report_progress (0, 1, "Writing");

   Then hdr should be passed NULL on subsequent calls while the operation is
   progressing. Once the operation is complete, a final call should be made as
   such to ensure proper termination of the progress report:

     report_progress (1, 1, NULL);

   It would be nice if we could reduce the usage to one and only one call for
   each of start, during and end cases. As things stand now, that is not
   possible and makes maintenance a bit more work. */

void report_progress (int completed, int total, char *hdr)
{
  static int last = 0;
  static double start_time;
  int percent = (completed * 100) / total;
  struct timeval tv;
  double t;

  if (update_progress == NULL)
    //return;
    update_progress = update_progress_tty;

  gettimeofday(&tv, NULL);
  t = tv.tv_sec + ((double)tv.tv_usec)/1000000;

  if (hdr) {
    last = 0;
    start_time = t;
    update_progress (percent, t - start_time, hdr);
  }

  if (percent > 100)
    percent = 100;

  if (percent > last) {
    last = percent;
    update_progress (percent, t - start_time, hdr);
  }

  if (percent == 100)
    last = 0;                   /* Get ready for next time. */
}

static void update_progress_tty (int percent, double etime, char *hdr)
{
  static char hashes[51];
  static char *header;
  static int last = 0;
  int i;

  hashes[50] = 0;

  memset (hashes, ' ', 50);
  for (i=0; i<percent; i+=2) {
    hashes[i/2] = '#';
  }

  if (hdr) {
    fprintf (stderr, "\n");
    last = 0;
    header = hdr;
  }

  if (last == 0) {
    fprintf(stderr, "\r%s | %s | %d%% %0.2fs", 
            header, hashes, percent, etime);
  }

  if (percent == 100) {
    last = 1;
    fprintf (stderr, "\n\n");
  }
}

static void update_progress_no_tty (int percent, double etime, char *hdr)
{
  static int done = 0;
  static int last = 0;
  int cnt = (percent>>1)*2;

  if (hdr) {
    fprintf (stderr, "\n%s | ", hdr);
    last = 0;
    done = 0;
  }
  else {
    while ((cnt > last) && (done == 0)) {
      fprintf (stderr, "#");
      cnt -=  2;
    }
  }

  if ((percent == 100) && (done == 0)) {
    fprintf (stderr, " | 100%% %0.2fs\n\n", etime);
    last = 0;
    done = 1;
  }
  else
    last = (percent>>1)*2;    /* Make last a multiple of 2. */
}

